import copy
import datetime
import uuid
from typing import Dict, List, Optional, Union

from marshmallow import Schema
from marshmallow import fields as schema_fields

from commercetools import types
from commercetools._schemas._common import ImageSchema, PriceSchema
from commercetools._schemas._product import (
    AttributeSchema,
    ProductDraftSchema,
    ProductPagedQueryResponseSchema,
    ProductSchema,
    ProductUpdateSchema,
    ProductVariantSchema,
)
from commercetools.services.products import _ProductQuerySchema
from commercetools.testing import utils
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import (
    create_commercetools_response,
    custom_fields_from_draft,
    get_product_variants,
    parse_request_params,
)


class ProductsModel(BaseModel):
    _primary_type_name = "product"
    _resource_schema = ProductSchema
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.ProductDraft, id: Optional[str] = None
    ) -> types.Product:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())

        master_variant = None
        if draft.master_variant:
            master_variant = self._create_variant_from_draft(draft.master_variant)

        product_data = types.ProductData(
            name=draft.name,
            categories=draft.categories,
            category_order_hints=draft.category_order_hints,
            description=draft.description,
            master_variant=master_variant,
            variants=[self._create_variant_from_draft(vd) for vd in draft.variants]
            if draft.variants
            else [],
            slug=draft.slug or types.LocalizedString(),
            search_keywords=types.SearchKeywords(),
        )

        if draft.publish:
            master_data = types.ProductCatalogData(
                staged=None,
                current=product_data,
                published=True,
                has_staged_changes=False,
            )
        else:
            master_data = types.ProductCatalogData(
                staged=product_data,
                current=None,
                published=False,
                has_staged_changes=True,
            )

        product = types.Product(
            id=str(object_id),
            key=draft.key,
            product_type=draft.product_type,
            master_data=master_data,
            version=1,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
        )

        return product

    def _create_variant_from_draft(
        self, draft: types.ProductVariantDraft
    ) -> types.ProductVariant:

        assets: Optional[List[types.Asset]] = None
        if draft.assets:
            assets = self._create_assets_from_draft(draft.assets)

        prices: Optional[List[types.Price]] = None
        if draft.prices:
            prices = self._create_prices_from_draft(draft.prices)

        return types.ProductVariant(
            id=1,
            sku=draft.sku,
            key=draft.key,
            prices=prices,
            attributes=draft.attributes,
            price=None,
            images=draft.images,
            assets=assets,
            availability=None,
            is_matching_variant=None,
            scoped_price=None,
            scoped_price_discounted=None,
        )

    def _create_assets_from_draft(
        self, drafts: List[types.AssetDraft]
    ) -> List[types.Asset]:
        assets: List[types.Asset] = []
        for draft in drafts:
            custom: Optional[types.CustomFields] = None
            if draft.custom:
                custom = custom_fields_from_draft(self._storage, draft.custom)

            asset = types.Asset(
                id=str(uuid.uuid4()),
                sources=draft.sources,
                name=draft.name,
                description=draft.description,
                tags=draft.tags,
                custom=custom,
                key=draft.key,
            )
            assets.append(asset)
        return assets

    def _create_prices_from_draft(
        self, drafts: List[types.PriceDraft]
    ) -> List[types.Price]:
        prices: List[types.Price] = []
        for draft in drafts:
            custom = None
            if draft.custom:
                custom = custom_fields_from_draft(self._storage, draft.custom)

            price = types.Price(
                id=str(uuid.uuid4()),
                value=self._create_price_from_draft(draft.value),
                country=draft.country,
                customer_group=draft.customer_group,
                channel=draft.channel,
                valid_from=draft.valid_from,
                valid_until=draft.valid_until,
                discounted=None,
                custom=custom,
                tiers=draft.tiers,
            )
            prices.append(price)
        return prices

    def _create_price_from_draft(
        self, draft: Union[types.Money, Optional[types.TypedMoneyDraft]]
    ) -> Optional[types.TypedMoney]:
        if draft is None:
            return None

        if isinstance(draft, types.CentPrecisionMoneyDraft):
            return types.CentPrecisionMoney(
                cent_amount=draft.cent_amount, currency_code=draft.currency_code
            )
        elif isinstance(draft, types.HighPrecisionMoneyDraft):
            return types.HighPrecisionMoney(
                cent_amount=draft.cent_amount,
                currency_code=draft.currency_code,
                precise_amount=draft.precise_amount,
            )
        elif isinstance(draft, types.Money):
            return types.CentPrecisionMoney(
                cent_amount=draft.cent_amount,
                currency_code=draft.currency_code,
                fraction_digits=2,
            )
        else:
            return types.TypedMoney(
                cent_amount=draft.cent_amount,
                currency_code=draft.currency_code,
                type=draft.type,
            )


def _get_target_obj(obj: dict, staged: bool):
    if not staged and obj["masterData"]["current"]:
        return obj["masterData"]["current"]
    return obj["masterData"]["staged"]


def _get_variant(data_object: dict, *, id: int = "", sku: str = "") -> Optional[dict]:
    if not data_object:
        return None

    variants = [data_object["masterVariant"]]
    if data_object["variants"]:
        variants += data_object["variants"]

    for variant in variants:
        if id and id == variant["id"]:
            return variant
        if sku and sku == variant["sku"]:
            return variant

    return None


def _update_productdata_attr(dst: str, src: str):
    def updater(self, obj: dict, action: types.ProductUpdateAction):
        value = getattr(action, src)
        target_obj = _get_target_obj(obj, getattr(action, "staged", False))

        if target_obj[dst] != value:
            new = copy.deepcopy(obj)
            new[dst] = value
            return new
        return obj

    return updater


def _set_attribute_action():
    def updater(self, obj: dict, action: types.ProductSetAttributeAction):
        staged = getattr(action, "staged", False)
        new = copy.deepcopy(obj)

        target_obj = _get_target_obj(new, staged)
        variant = _get_variant(target_obj, id=action.variant_id)
        if variant:
            for attr in variant["attributes"]:
                if attr["name"] == action.name:
                    attr["value"] = action.value
                    return new

            # Attribute not found, will add it
            attr_schema = AttributeSchema()
            variant["attributes"].append(
                attr_schema.dump(types.Attribute(name=action.name, value=action.value))
            )

        return new

    return updater


def _add_variant_action():
    def updater(self, obj: dict, action: types.ProductAddVariantAction):
        variant = self.model._create_variant_from_draft(
            types.ProductVariantDraft(
                sku=action.sku,
                key=action.key,
                prices=action.prices,
                attributes=action.attributes,
                images=action.images,
                assets=action.assets,
            )
        )
        schema = ProductVariantSchema()

        new = copy.deepcopy(obj)
        target_obj = _get_target_obj(new, staged=getattr(action, "staged", True))
        if not target_obj["variants"]:
            target_obj["variants"] = []
        target_obj["variants"].append(schema.dump(variant))

        return new

    return updater


def _publish_product_action():
    def updater(self, obj: dict, action: types.ProductPublishAction):
        new = copy.deepcopy(obj)
        # not implemented scopes right now.
        if new["masterData"].get("staged"):
            new["masterData"]["current"] = new["masterData"]["staged"]
            new["masterData"]["staged"] = None
        new["masterData"]["hasStagedChanges"] = False
        new["masterData"]["published"] = True
        return new

    return updater


def convert_draft_price(
    price_draft: types.PriceDraft, price_id: str = None
) -> types.Price:
    tiers: Optional[List[types.PriceTier]] = None
    if price_draft.tiers:
        tiers = [utils.create_from_draft(tier) for tier in price_draft.tiers]
    return types.Price(
        id=price_id or str(uuid.uuid4()),
        country=price_draft.country,
        channel=price_draft.channel,
        value=utils._money_to_typed(price_draft.value),
        valid_from=price_draft.valid_from,
        valid_until=price_draft.valid_until,
        discounted=price_draft.discounted,
        custom=utils.create_from_draft(price_draft.custom),
        tiers=tiers,
    )


def _set_product_prices():
    def updater(self, obj: dict, action: types.ProductSetPricesAction):
        new = copy.deepcopy(obj)
        target_obj = _get_target_obj(new, getattr(action, "staged", True))
        prices = []
        for price_draft in action.prices:
            price = convert_draft_price(price_draft)
            prices.append(price)

        schema = PriceSchema()
        for variant in get_product_variants(target_obj):
            if variant["sku"] == action.sku:
                variant["prices"] = schema.dump(prices, many=True)
        if action.staged:
            new["masterData"]["hasStagedChanges"] = True
        return new

    return updater


def _change_price():
    def updater(self, obj: dict, action: types.ProductChangePriceAction):
        new = copy.deepcopy(obj)
        staged = action.staged
        if staged is None:
            staged = True
        target_obj = _get_target_obj(new, staged)
        changed_price = convert_draft_price(action.price, action.price_id)
        schema = PriceSchema()

        found_price = True
        for variant in get_product_variants(target_obj):
            if not variant["prices"]:
                continue
            for index, price in enumerate(variant["prices"]):
                if price["id"] == action.price_id:
                    variant["prices"][index] = schema.dump(changed_price)
                    found_price = True
                    break
        if not found_price:
            raise ValueError("Could not find price with id %s" % action.price_id)
        if staged:
            new["masterData"]["hasStagedChanges"] = True
        return new

    return updater


def _add_price():
    def updater(self, obj: dict, action: types.ProductAddPriceAction):
        new = copy.deepcopy(obj)
        staged = action.staged
        if staged is None:
            staged = True
        target_obj = _get_target_obj(new, staged)
        new_price = convert_draft_price(action.price)
        schema = PriceSchema()

        found_sku = False
        for variant in get_product_variants(target_obj):
            if variant["sku"] == action.sku:
                if "prices" not in variant:
                    variant["prices"] = []
                elif not variant["prices"]:
                    variant["prices"] = []
                variant["prices"].append(schema.dump(new_price))
                found_sku = True
                break
        if not found_sku:
            raise ValueError("Could not find sku %s" % action.sku)
        if staged:
            new["masterData"]["hasStagedChanges"] = True
        return new

    return updater


def _change_master_variant():
    def updater(self, obj: Dict, action: types.ProductChangeMasterVariantAction):
        new = copy.deepcopy(obj)
        staged = action.staged
        if staged is None:
            staged = True
        target_obj = _get_target_obj(new, staged)
        master_variant = target_obj["masterVariant"]
        if master_variant and master_variant["sku"] == action.sku:
            return obj
        found_sku = False
        for variant in get_product_variants(target_obj):
            if variant["sku"] == action.sku:
                target_obj["variants"].append(target_obj["masterVariant"])
                target_obj["masterVariant"] = variant
                target_obj["variants"].remove(variant)
                found_sku = True
                break
        if not found_sku:
            raise ValueError("Could not find sku %s" % action.sku)
        if staged:
            new["masterData"]["hasStagedChanges"] = True
        return new

    return updater


class UploadImageQuerySchema(Schema):
    staged = schema_fields.Bool()
    filename = schema_fields.Field()
    sku = schema_fields.Field()


class ProductsBackend(ServiceBackend):
    service_path = "products"
    model_class = ProductsModel
    _schema_draft = ProductDraftSchema
    _schema_update = ProductUpdateSchema
    _schema_query_params = _ProductQuerySchema
    _schema_query_response = ProductPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
            ("^(?P<id>[^/]+)\/images$", "POST", self.upload_image),
        ]

    # Fixme: use decorator for this
    _actions = {
        "changeSlug": _update_productdata_attr("slug", "slug"),
        "setAttribute": _set_attribute_action(),
        "addVariant": _add_variant_action(),
        "setPrices": _set_product_prices(),
        "changePrice": _change_price(),
        "addPrice": _add_price(),
        "changeMasterVariant": _change_master_variant(),
        "publish": _publish_product_action(),
    }

    def upload_image(self, request, id):
        obj = self.model.get_by_id(id)
        if not obj:
            return create_commercetools_response(request, status_code=404)

        obj = copy.deepcopy(obj)

        params = parse_request_params(UploadImageQuerySchema, request)
        target = _get_target_obj(obj, staged=params["staged"])
        filename = params["filename"]

        variant = _get_variant(target, sku=params["sku"])
        if not variant["images"]:
            variant["images"] = []
        image_schema = ImageSchema()
        variant["images"].append(
            image_schema.dump(
                types.Image(
                    url=f"cdn.commercetools.local/detail-{filename}",
                    dimensions=types.ImageDimensions(w=500, h=500),
                )
            )
        )
        self.model.save(obj)
        return create_commercetools_response(request, json=obj)



# DO NOT EDIT! This file is automatically generated
import marshmallow
import marshmallow_enum

from commercetools import helpers, types

__all__ = [
    "AddressSchema",
    "AssetDimensionsSchema",
    "AssetDraftSchema",
    "AssetSchema",
    "AssetSourceSchema",
    "BaseResourceSchema",
    "CentPrecisionMoneyDraftSchema",
    "CentPrecisionMoneySchema",
    "ClientLoggingSchema",
    "CreatedBySchema",
    "DiscountedPriceSchema",
    "GeoJsonPointSchema",
    "GeoJsonSchema",
    "HighPrecisionMoneyDraftSchema",
    "HighPrecisionMoneySchema",
    "ImageDimensionsSchema",
    "ImageSchema",
    "KeyReferenceSchema",
    "LastModifiedBySchema",
    "MoneySchema",
    "PagedQueryResponseSchema",
    "PriceDraftSchema",
    "PriceSchema",
    "PriceTierDraftSchema",
    "PriceTierSchema",
    "QueryPriceSchema",
    "ReferenceSchema",
    "ResourceIdentifierSchema",
    "ScopedPriceSchema",
    "TypedMoneyDraftSchema",
    "TypedMoneySchema",
    "UpdateActionSchema",
    "UpdateSchema",
]


class LocalizedStringField(marshmallow.fields.Dict):
    def _deserialize(self, value, attr, data, **kwargs):
        result = super()._deserialize(value, attr, data)
        return types.LocalizedString(**result)


class AddressSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Address`."""

    id = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    title = marshmallow.fields.String(allow_none=True, missing=None)
    salutation = marshmallow.fields.String(allow_none=True, missing=None)
    first_name = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="firstName"
    )
    last_name = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="lastName"
    )
    street_name = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="streetName"
    )
    street_number = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="streetNumber"
    )
    additional_street_info = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="additionalStreetInfo"
    )
    postal_code = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="postalCode"
    )
    city = marshmallow.fields.String(allow_none=True, missing=None)
    region = marshmallow.fields.String(allow_none=True, missing=None)
    state = marshmallow.fields.String(allow_none=True, missing=None)
    country = marshmallow.fields.String()
    company = marshmallow.fields.String(allow_none=True, missing=None)
    department = marshmallow.fields.String(allow_none=True, missing=None)
    building = marshmallow.fields.String(allow_none=True, missing=None)
    apartment = marshmallow.fields.String(allow_none=True, missing=None)
    p_o_box = marshmallow.fields.String(allow_none=True, missing=None, data_key="pOBox")
    phone = marshmallow.fields.String(allow_none=True, missing=None)
    mobile = marshmallow.fields.String(allow_none=True, missing=None)
    email = marshmallow.fields.String(allow_none=True, missing=None)
    fax = marshmallow.fields.String(allow_none=True, missing=None)
    additional_address_info = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="additionalAddressInfo"
    )
    external_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="externalId"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Address(**data)


class AssetDimensionsSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.AssetDimensions`."""

    w = marshmallow.fields.Integer(allow_none=True)
    h = marshmallow.fields.Integer(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.AssetDimensions(**data)


class AssetDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.AssetDraft`."""

    sources = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetSourceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    name = LocalizedStringField(allow_none=True)
    description = LocalizedStringField(allow_none=True, missing=None)
    tags = marshmallow.fields.List(
        marshmallow.fields.String(allow_none=True), missing=None
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    key = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.AssetDraft(**data)


class AssetSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Asset`."""

    id = marshmallow.fields.String(allow_none=True)
    sources = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetSourceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    name = LocalizedStringField(allow_none=True)
    description = LocalizedStringField(allow_none=True, missing=None)
    tags = marshmallow.fields.List(
        marshmallow.fields.String(allow_none=True), missing=None
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    key = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Asset(**data)


class AssetSourceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.AssetSource`."""

    uri = marshmallow.fields.String(allow_none=True)
    key = marshmallow.fields.String(allow_none=True, missing=None)
    dimensions = helpers.LazyNestedField(
        nested="commercetools._schemas._common.AssetDimensionsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    content_type = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="contentType"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.AssetSource(**data)


class BaseResourceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.BaseResource`."""

    id = marshmallow.fields.String(allow_none=True)
    version = marshmallow.fields.Integer(allow_none=True)
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")
    last_modified_at = marshmallow.fields.DateTime(
        allow_none=True, data_key="lastModifiedAt"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.BaseResource(**data)


class ClientLoggingSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ClientLogging`."""

    client_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="clientId"
    )
    external_user_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="externalUserId"
    )
    customer = helpers.LazyNestedField(
        nested="commercetools._schemas._customer.CustomerReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    anonymous_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="anonymousId"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ClientLogging(**data)


class DiscountedPriceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.DiscountedPrice`."""

    value = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    discount = helpers.LazyNestedField(
        nested="commercetools._schemas._product_discount.ProductDiscountReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.DiscountedPrice(**data)


class GeoJsonSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.GeoJson`."""

    type = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.GeoJson(**data)


class ImageDimensionsSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ImageDimensions`."""

    w = marshmallow.fields.Integer(allow_none=True)
    h = marshmallow.fields.Integer(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ImageDimensions(**data)


class ImageSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Image`."""

    url = marshmallow.fields.String(allow_none=True)
    dimensions = helpers.LazyNestedField(
        nested="commercetools._schemas._common.ImageDimensionsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    label = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Image(**data)


class KeyReferenceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.KeyReference`."""

    type_id = marshmallow_enum.EnumField(
        types.ReferenceTypeId, by_value=True, data_key="typeId"
    )
    key = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.KeyReference(**data)


class MoneySchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Money`."""

    cent_amount = marshmallow.fields.Integer(allow_none=True, data_key="centAmount")
    currency_code = marshmallow.fields.String(data_key="currencyCode")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Money(**data)


class PagedQueryResponseSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.PagedQueryResponse`."""

    limit = marshmallow.fields.Integer(allow_none=True)
    count = marshmallow.fields.Integer(allow_none=True)
    total = marshmallow.fields.Integer(allow_none=True, missing=None)
    offset = marshmallow.fields.Integer(allow_none=True)
    results = helpers.LazyNestedField(
        nested="commercetools._schemas._common.BaseResourceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    facets = helpers.LazyNestedField(
        nested="commercetools._schemas._product.FacetResultsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    meta = marshmallow.fields.Dict(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.PagedQueryResponse(**data)


class PriceDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.PriceDraft`."""

    value = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    country = marshmallow.fields.String(missing=None)
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelResourceIdentifierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    valid_from = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validFrom"
    )
    valid_until = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validUntil"
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    tiers = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceTierDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )
    discounted = helpers.LazyNestedField(
        nested="commercetools._schemas._common.DiscountedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.PriceDraft(**data)


class PriceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Price`."""

    id = marshmallow.fields.String(allow_none=True)
    value = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "centPrecision": "commercetools._schemas._common.CentPrecisionMoneySchema",
            "highPrecision": "commercetools._schemas._common.HighPrecisionMoneySchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    country = marshmallow.fields.String(missing=None)
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    valid_from = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validFrom"
    )
    valid_until = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validUntil"
    )
    discounted = helpers.LazyNestedField(
        nested="commercetools._schemas._common.DiscountedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    tiers = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceTierSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Price(**data)


class PriceTierDraftSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.PriceTierDraft`."""

    minimum_quantity = marshmallow.fields.Integer(
        allow_none=True, data_key="minimumQuantity"
    )
    value = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.PriceTierDraft(**data)


class PriceTierSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.PriceTier`."""

    minimum_quantity = marshmallow.fields.Integer(
        allow_none=True, data_key="minimumQuantity"
    )
    value = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "centPrecision": "commercetools._schemas._common.CentPrecisionMoneySchema",
            "highPrecision": "commercetools._schemas._common.HighPrecisionMoneySchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.PriceTier(**data)


class QueryPriceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.QueryPrice`."""

    id = marshmallow.fields.String(allow_none=True)
    value = helpers.LazyNestedField(
        nested="commercetools._schemas._common.MoneySchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    country = marshmallow.fields.String(missing=None)
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    valid_from = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validFrom"
    )
    valid_until = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validUntil"
    )
    discounted = helpers.LazyNestedField(
        nested="commercetools._schemas._common.DiscountedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    tiers = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceTierDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.QueryPrice(**data)


class ReferenceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Reference`."""

    type_id = marshmallow_enum.EnumField(
        types.ReferenceTypeId, by_value=True, data_key="typeId"
    )
    id = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.Reference(**data)


class ResourceIdentifierSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ResourceIdentifier`."""

    type_id = marshmallow_enum.EnumField(
        types.ReferenceTypeId, by_value=True, missing=None, data_key="typeId"
    )
    id = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type_id"]
        return types.ResourceIdentifier(**data)


class ScopedPriceSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ScopedPrice`."""

    id = marshmallow.fields.String(allow_none=True)
    value = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "centPrecision": "commercetools._schemas._common.CentPrecisionMoneySchema",
            "highPrecision": "commercetools._schemas._common.HighPrecisionMoneySchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    current_value = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "centPrecision": "commercetools._schemas._common.CentPrecisionMoneySchema",
            "highPrecision": "commercetools._schemas._common.HighPrecisionMoneySchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="currentValue",
    )
    country = marshmallow.fields.String(missing=None)
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    valid_from = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validFrom"
    )
    valid_until = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validUntil"
    )
    discounted = helpers.LazyNestedField(
        nested="commercetools._schemas._common.DiscountedPriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )
    custom = helpers.LazyNestedField(
        nested="commercetools._schemas._type.CustomFieldsSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ScopedPrice(**data)


class TypedMoneySchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.TypedMoney`."""

    type = marshmallow_enum.EnumField(types.MoneyType, by_value=True)
    fraction_digits = marshmallow.fields.Integer(
        allow_none=True, data_key="fractionDigits"
    )
    cent_amount = marshmallow.fields.Integer(allow_none=True, data_key="centAmount")
    currency_code = marshmallow.fields.String(data_key="currencyCode")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.TypedMoney(**data)


class UpdateActionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.UpdateAction`."""

    action = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.UpdateAction(**data)


class UpdateSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Update`."""

    version = marshmallow.fields.Integer(allow_none=True)
    actions = marshmallow.fields.List(
        helpers.LazyNestedField(
            nested="commercetools._schemas._common.UpdateActionSchema",
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Update(**data)


class CentPrecisionMoneySchema(TypedMoneySchema):
    """Marshmallow schema for :class:`commercetools.types.CentPrecisionMoney`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CentPrecisionMoney(**data)


class CreatedBySchema(ClientLoggingSchema):
    """Marshmallow schema for :class:`commercetools.types.CreatedBy`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.CreatedBy(**data)


class GeoJsonPointSchema(GeoJsonSchema):
    """Marshmallow schema for :class:`commercetools.types.GeoJsonPoint`."""

    coordinates = marshmallow.fields.List(
        marshmallow.fields.Integer(allow_none=True), allow_none=True
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.GeoJsonPoint(**data)


class HighPrecisionMoneySchema(TypedMoneySchema):
    """Marshmallow schema for :class:`commercetools.types.HighPrecisionMoney`."""

    precise_amount = marshmallow.fields.Integer(
        allow_none=True, data_key="preciseAmount"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.HighPrecisionMoney(**data)


class LastModifiedBySchema(ClientLoggingSchema):
    """Marshmallow schema for :class:`commercetools.types.LastModifiedBy`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.LastModifiedBy(**data)


class TypedMoneyDraftSchema(MoneySchema):
    """Marshmallow schema for :class:`commercetools.types.TypedMoneyDraft`."""

    type = marshmallow_enum.EnumField(types.MoneyType, by_value=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.TypedMoneyDraft(**data)


class CentPrecisionMoneyDraftSchema(TypedMoneyDraftSchema):
    """Marshmallow schema for :class:`commercetools.types.CentPrecisionMoneyDraft`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CentPrecisionMoneyDraft(**data)


class HighPrecisionMoneyDraftSchema(TypedMoneyDraftSchema):
    """Marshmallow schema for :class:`commercetools.types.HighPrecisionMoneyDraft`."""

    precise_amount = marshmallow.fields.Integer(
        allow_none=True, data_key="preciseAmount"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.HighPrecisionMoneyDraft(**data)



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import (
    BaseResource,
    Reference,
    ReferenceTypeId,
    ResourceIdentifier,
)

if typing.TYPE_CHECKING:
    from ._common import CreatedBy, LastModifiedBy
    from ._customer import CustomerReference, CustomerResourceIdentifier
    from ._product import ProductReference, ProductResourceIdentifier
    from ._state import StateReference, StateResourceIdentifier
    from ._type import (
        CustomFields,
        CustomFieldsDraft,
        FieldContainer,
        TypeResourceIdentifier,
    )
__all__ = [
    "Review",
    "ReviewDraft",
    "ReviewPagedQueryResponse",
    "ReviewRatingStatistics",
    "ReviewReference",
    "ReviewResourceIdentifier",
    "ReviewSetAuthorNameAction",
    "ReviewSetCustomFieldAction",
    "ReviewSetCustomTypeAction",
    "ReviewSetCustomerAction",
    "ReviewSetKeyAction",
    "ReviewSetLocaleAction",
    "ReviewSetRatingAction",
    "ReviewSetTargetAction",
    "ReviewSetTextAction",
    "ReviewSetTitleAction",
    "ReviewTransitionStateAction",
    "ReviewUpdate",
    "ReviewUpdateAction",
]


class Review(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``uniquenessValue`` `in Commercetools)`
    uniqueness_value: typing.Optional[str]
    #: Optional :class:`str`
    locale: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``authorName`` `in Commercetools)`
    author_name: typing.Optional[str]
    #: Optional :class:`str`
    title: typing.Optional[str]
    #: Optional :class:`str`
    text: typing.Optional[str]
    #: Optional :class:`commercetools.types.ProductReference`
    target: typing.Optional["ProductReference"]
    #: :class:`bool` `(Named` ``includedInStatistics`` `in Commercetools)`
    included_in_statistics: bool
    #: Optional :class:`int`
    rating: typing.Optional[int]
    #: Optional :class:`commercetools.types.StateReference`
    state: typing.Optional["StateReference"]
    #: Optional :class:`commercetools.types.CustomerReference`
    customer: typing.Optional["CustomerReference"]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        included_in_statistics: bool,
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None,
        key: typing.Optional[str] = None,
        uniqueness_value: typing.Optional[str] = None,
        locale: typing.Optional[str] = None,
        author_name: typing.Optional[str] = None,
        title: typing.Optional[str] = None,
        text: typing.Optional[str] = None,
        target: typing.Optional["ProductReference"] = None,
        rating: typing.Optional[int] = None,
        state: typing.Optional["StateReference"] = None,
        customer: typing.Optional["CustomerReference"] = None,
        custom: typing.Optional["CustomFields"] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.key = key
        self.uniqueness_value = uniqueness_value
        self.locale = locale
        self.author_name = author_name
        self.title = title
        self.text = text
        self.target = target
        self.included_in_statistics = included_in_statistics
        self.rating = rating
        self.state = state
        self.customer = customer
        self.custom = custom
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "Review(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, key=%r, uniqueness_value=%r, locale=%r, author_name=%r, title=%r, text=%r, target=%r, included_in_statistics=%r, rating=%r, state=%r, customer=%r, custom=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.key,
                self.uniqueness_value,
                self.locale,
                self.author_name,
                self.title,
                self.text,
                self.target,
                self.included_in_statistics,
                self.rating,
                self.state,
                self.customer,
                self.custom,
            )
        )


class ReviewDraft(_BaseType):
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``uniquenessValue`` `in Commercetools)`
    uniqueness_value: typing.Optional[str]
    #: Optional :class:`str`
    locale: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``authorName`` `in Commercetools)`
    author_name: typing.Optional[str]
    #: Optional :class:`str`
    title: typing.Optional[str]
    #: Optional :class:`str`
    text: typing.Optional[str]
    #: Optional :class:`commercetools.types.ProductResourceIdentifier`
    target: typing.Optional["ProductResourceIdentifier"]
    #: Optional :class:`commercetools.types.StateResourceIdentifier`
    state: typing.Optional["StateResourceIdentifier"]
    #: Optional :class:`int`
    rating: typing.Optional[int]
    #: Optional :class:`commercetools.types.CustomerResourceIdentifier`
    customer: typing.Optional["CustomerResourceIdentifier"]
    #: Optional :class:`commercetools.types.CustomFieldsDraft`
    custom: typing.Optional["CustomFieldsDraft"]

    def __init__(
        self,
        *,
        key: typing.Optional[str] = None,
        uniqueness_value: typing.Optional[str] = None,
        locale: typing.Optional[str] = None,
        author_name: typing.Optional[str] = None,
        title: typing.Optional[str] = None,
        text: typing.Optional[str] = None,
        target: typing.Optional["ProductResourceIdentifier"] = None,
        state: typing.Optional["StateResourceIdentifier"] = None,
        rating: typing.Optional[int] = None,
        customer: typing.Optional["CustomerResourceIdentifier"] = None,
        custom: typing.Optional["CustomFieldsDraft"] = None
    ) -> None:
        self.key = key
        self.uniqueness_value = uniqueness_value
        self.locale = locale
        self.author_name = author_name
        self.title = title
        self.text = text
        self.target = target
        self.state = state
        self.rating = rating
        self.customer = customer
        self.custom = custom
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ReviewDraft(key=%r, uniqueness_value=%r, locale=%r, author_name=%r, title=%r, text=%r, target=%r, state=%r, rating=%r, customer=%r, custom=%r)"
            % (
                self.key,
                self.uniqueness_value,
                self.locale,
                self.author_name,
                self.title,
                self.text,
                self.target,
                self.state,
                self.rating,
                self.customer,
                self.custom,
            )
        )


class ReviewPagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.Review`
    results: typing.Sequence["Review"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["Review"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ReviewPagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class ReviewRatingStatistics(_BaseType):
    #: :class:`int` `(Named` ``averageRating`` `in Commercetools)`
    average_rating: int
    #: :class:`int` `(Named` ``highestRating`` `in Commercetools)`
    highest_rating: int
    #: :class:`int` `(Named` ``lowestRating`` `in Commercetools)`
    lowest_rating: int
    #: :class:`int`
    count: int
    #: :class:`object` `(Named` ``ratingsDistribution`` `in Commercetools)`
    ratings_distribution: object

    def __init__(
        self,
        *,
        average_rating: int,
        highest_rating: int,
        lowest_rating: int,
        count: int,
        ratings_distribution: object
    ) -> None:
        self.average_rating = average_rating
        self.highest_rating = highest_rating
        self.lowest_rating = lowest_rating
        self.count = count
        self.ratings_distribution = ratings_distribution
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ReviewRatingStatistics(average_rating=%r, highest_rating=%r, lowest_rating=%r, count=%r, ratings_distribution=%r)"
            % (
                self.average_rating,
                self.highest_rating,
                self.lowest_rating,
                self.count,
                self.ratings_distribution,
            )
        )


class ReviewReference(Reference):
    #: Optional :class:`commercetools.types.Review`
    obj: typing.Optional["Review"]

    def __init__(self, *, id: str, obj: typing.Optional["Review"] = None) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.REVIEW, id=id)

    def __repr__(self) -> str:
        return "ReviewReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )


class ReviewResourceIdentifier(ResourceIdentifier):
    def __init__(
        self, *, id: typing.Optional[str] = None, key: typing.Optional[str] = None
    ) -> None:
        super().__init__(type_id=ReferenceTypeId.REVIEW, id=id, key=key)

    def __repr__(self) -> str:
        return "ReviewResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class ReviewUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "ReviewUpdate(version=%r, actions=%r)" % (self.version, self.actions)


class ReviewUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "ReviewUpdateAction(action=%r)" % (self.action,)


class ReviewSetAuthorNameAction(ReviewUpdateAction):
    #: Optional :class:`str` `(Named` ``authorName`` `in Commercetools)`
    author_name: typing.Optional[str]

    def __init__(self, *, author_name: typing.Optional[str] = None) -> None:
        self.author_name = author_name
        super().__init__(action="setAuthorName")

    def __repr__(self) -> str:
        return "ReviewSetAuthorNameAction(action=%r, author_name=%r)" % (
            self.action,
            self.author_name,
        )


class ReviewSetCustomFieldAction(ReviewUpdateAction):
    #: :class:`str`
    name: str
    #: Optional :class:`typing.Any`
    value: typing.Optional[typing.Any]

    def __init__(self, *, name: str, value: typing.Optional[typing.Any] = None) -> None:
        self.name = name
        self.value = value
        super().__init__(action="setCustomField")

    def __repr__(self) -> str:
        return "ReviewSetCustomFieldAction(action=%r, name=%r, value=%r)" % (
            self.action,
            self.name,
            self.value,
        )


class ReviewSetCustomTypeAction(ReviewUpdateAction):
    #: Optional :class:`commercetools.types.TypeResourceIdentifier`
    type: typing.Optional["TypeResourceIdentifier"]
    #: Optional :class:`commercetools.types.FieldContainer`
    fields: typing.Optional["FieldContainer"]

    def __init__(
        self,
        *,
        type: typing.Optional["TypeResourceIdentifier"] = None,
        fields: typing.Optional["FieldContainer"] = None
    ) -> None:
        self.type = type
        self.fields = fields
        super().__init__(action="setCustomType")

    def __repr__(self) -> str:
        return "ReviewSetCustomTypeAction(action=%r, type=%r, fields=%r)" % (
            self.action,
            self.type,
            self.fields,
        )


class ReviewSetCustomerAction(ReviewUpdateAction):
    #: Optional :class:`commercetools.types.CustomerResourceIdentifier`
    customer: typing.Optional["CustomerResourceIdentifier"]

    def __init__(
        self, *, customer: typing.Optional["CustomerResourceIdentifier"] = None
    ) -> None:
        self.customer = customer
        super().__init__(action="setCustomer")

    def __repr__(self) -> str:
        return "ReviewSetCustomerAction(action=%r, customer=%r)" % (
            self.action,
            self.customer,
        )


class ReviewSetKeyAction(ReviewUpdateAction):
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(self, *, key: typing.Optional[str] = None) -> None:
        self.key = key
        super().__init__(action="setKey")

    def __repr__(self) -> str:
        return "ReviewSetKeyAction(action=%r, key=%r)" % (self.action, self.key)


class ReviewSetLocaleAction(ReviewUpdateAction):
    #: Optional :class:`str`
    locale: typing.Optional[str]

    def __init__(self, *, locale: typing.Optional[str] = None) -> None:
        self.locale = locale
        super().__init__(action="setLocale")

    def __repr__(self) -> str:
        return "ReviewSetLocaleAction(action=%r, locale=%r)" % (
            self.action,
            self.locale,
        )


class ReviewSetRatingAction(ReviewUpdateAction):
    #: Optional :class:`int`
    rating: typing.Optional[int]

    def __init__(self, *, rating: typing.Optional[int] = None) -> None:
        self.rating = rating
        super().__init__(action="setRating")

    def __repr__(self) -> str:
        return "ReviewSetRatingAction(action=%r, rating=%r)" % (
            self.action,
            self.rating,
        )


class ReviewSetTargetAction(ReviewUpdateAction):
    #: :class:`commercetools.types.ProductResourceIdentifier`
    target: "ProductResourceIdentifier"

    def __init__(self, *, target: "ProductResourceIdentifier") -> None:
        self.target = target
        super().__init__(action="setTarget")

    def __repr__(self) -> str:
        return "ReviewSetTargetAction(action=%r, target=%r)" % (
            self.action,
            self.target,
        )


class ReviewSetTextAction(ReviewUpdateAction):
    #: Optional :class:`str`
    text: typing.Optional[str]

    def __init__(self, *, text: typing.Optional[str] = None) -> None:
        self.text = text
        super().__init__(action="setText")

    def __repr__(self) -> str:
        return "ReviewSetTextAction(action=%r, text=%r)" % (self.action, self.text)


class ReviewSetTitleAction(ReviewUpdateAction):
    #: Optional :class:`str`
    title: typing.Optional[str]

    def __init__(self, *, title: typing.Optional[str] = None) -> None:
        self.title = title
        super().__init__(action="setTitle")

    def __repr__(self) -> str:
        return "ReviewSetTitleAction(action=%r, title=%r)" % (self.action, self.title)


class ReviewTransitionStateAction(ReviewUpdateAction):
    #: :class:`commercetools.types.StateResourceIdentifier`
    state: "StateResourceIdentifier"
    #: Optional :class:`bool`
    force: typing.Optional[bool]

    def __init__(
        self, *, state: "StateResourceIdentifier", force: typing.Optional[bool] = None
    ) -> None:
        self.state = state
        self.force = force
        super().__init__(action="transitionState")

    def __repr__(self) -> str:
        return "ReviewTransitionStateAction(action=%r, state=%r, force=%r)" % (
            self.action,
            self.state,
            self.force,
        )



# DO NOT EDIT! This file is automatically generated
from ._api_client import *  # noqa
from ._cart import *  # noqa
from ._cart_discount import *  # noqa
from ._category import *  # noqa
from ._channel import *  # noqa
from ._common import *  # noqa
from ._custom_object import *  # noqa
from ._customer import *  # noqa
from ._customer_group import *  # noqa
from ._discount_code import *  # noqa
from ._error import *  # noqa
from ._extension import *  # noqa
from ._graph_ql import *  # noqa
from ._inventory import *  # noqa
from ._me import *  # noqa
from ._message import *  # noqa
from ._order import *  # noqa
from ._order_edit import *  # noqa
from ._payment import *  # noqa
from ._product import *  # noqa
from ._product_discount import *  # noqa
from ._product_type import *  # noqa
from ._project import *  # noqa
from ._review import *  # noqa
from ._shipping_method import *  # noqa
from ._shopping_list import *  # noqa
from ._state import *  # noqa
from ._store import *  # noqa
from ._subscription import *  # noqa
from ._tax_category import *  # noqa
from ._type import *  # noqa
from ._zone import *  # noqa



import pytest
from requests.exceptions import HTTPError

from commercetools import types


def test_channel_get_by_id(client):
    channel = client.channels.create(
        types.ChannelDraft(
            key="test-channel", roles=[types.ChannelRoleEnum.INVENTORY_SUPPLY]
        )
    )

    assert channel.id
    assert channel.key == "test-channel"

    channel = client.channels.get_by_id(channel.id)
    assert channel.id
    assert channel.key == "test-channel"

    with pytest.raises(HTTPError):
        client.channels.get_by_id("invalid")


def test_channel_query(client):
    client.channels.create(
        types.ChannelDraft(
            key="test-channel1", roles=[types.ChannelRoleEnum.INVENTORY_SUPPLY]
        )
    )
    client.channels.create(
        types.ChannelDraft(
            key="test-channel2", roles=[types.ChannelRoleEnum.INVENTORY_SUPPLY]
        )
    )

    # single sort query
    result = client.channels.query(sort="id asc")
    assert len(result.results) == 2
    assert result.total == 2

    # multiple sort queries
    result = client.channels.query(sort=["id asc", "name asc"])
    assert len(result.results) == 2
    assert result.total == 2


def test_channel_update(client):
    """Test the return value of the update methods.

    It doesn't test the actual update itself.
    TODO: See if this is worth testing since we're using a mocking backend
    """
    channel = client.channels.create(
        types.ChannelDraft(
            key="test-channel",
            name=types.LocalizedString(nl="nl-channel"),
            roles=[types.ChannelRoleEnum.INVENTORY_SUPPLY],
        )
    )
    assert channel.key == "test-channel"

    channel = client.channels.update_by_id(
        id=channel.id,
        version=channel.version,
        actions=[
            types.ChannelChangeNameAction(name=types.LocalizedString(nl="nl-channel2"))
        ],
    )



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._api_client import (
    ApiClientDraftSchema,
    ApiClientPagedQueryResponseSchema,
    ApiClientSchema,
)
from commercetools.testing.abstract import BaseModel, ServiceBackend


class ApiClientsModel(BaseModel):
    _primary_type_name = "api-client"
    _resource_schema = ApiClientSchema

    def _create_from_draft(
        self, draft: types.ApiClientDraft, id: typing.Optional[str] = None
    ) -> types.ApiClient:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.ApiClient(
            id=str(object_id),
            name=draft.name,
            scope=draft.scope,
            secret=str(uuid.uuid4()),
            created_at=datetime.datetime.now(datetime.timezone.utc),
        )


class ApiClientsBackend(ServiceBackend):
    service_path = "api-clients"
    model_class = ApiClientsModel
    _schema_draft = ApiClientDraftSchema
    _schema_query_response = ApiClientPagedQueryResponseSchema
    _verify_version = False

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]



from commercetools import types


def test_tax_category_create(client):
    tax_category = client.tax_categories.create(
        types.TaxCategoryDraft(name="Hoog", rates=[])
    )

    assert tax_category.id
    assert tax_category.name == "Hoog"


def test_tax_category_get_by_id(client):
    tax_category = client.tax_categories.create(
        types.TaxCategoryDraft(name="Hoog", rates=[])
    )

    assert tax_category.id
    assert tax_category.name == "Hoog"

    tax_category = client.tax_categories.get_by_id(tax_category.id)
    assert tax_category.id
    assert tax_category.name == "Hoog"


def test_tax_category_update_by_id(client):
    tax_category = client.tax_categories.create(
        types.TaxCategoryDraft(name="Hoog", rates=[])
    )

    assert tax_category.id
    assert tax_category.name == "Hoog"

    tax_category = client.tax_categories.update_by_id(
        tax_category.id,
        version=tax_category.version,
        actions=[types.TaxCategorySetDescriptionAction(description="Some text")],
    )
    assert tax_category.id
    assert tax_category.name == "Hoog"



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._cart_discount import (
    CartDiscountDraftSchema,
    CartDiscountPagedQueryResponseSchema,
    CartDiscountSchema,
    CartDiscountUpdateSchema,
)
from commercetools.testing import utils
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import update_attribute, update_enum_attribute


class CartDiscountsModel(BaseModel):
    _resource_schema = CartDiscountSchema
    _primary_type_name = "cart-discounts"
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.CartDiscountDraft, id: typing.Optional[str] = None
    ) -> types.CartDiscount:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.CartDiscount(
            id=str(object_id),
            version=1,
            key=draft.key,
            name=draft.name,
            description=draft.description,
            value=draft.value,
            target=draft.target,
            cart_predicate=draft.cart_predicate,
            is_active=draft.is_active or False,
            references=[],
            stacking_mode=draft.stacking_mode or types.StackingMode.STACKING,
            sort_order=draft.sort_order,
            valid_from=draft.valid_from,
            valid_until=draft.valid_until,
            requires_discount_code=draft.requires_discount_code,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            custom=utils.create_from_draft(draft.custom),
        )


class CartDiscountsBackend(ServiceBackend):
    service_path = "cart-discounts"
    model_class = CartDiscountsModel
    _schema_draft = CartDiscountDraftSchema
    _schema_update = CartDiscountUpdateSchema
    _schema_query_response = CartDiscountPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
        ]

    _actions = {
        "setKey": update_attribute("key", "key"),
        "changeSortOrder": update_attribute("sortOrder", "sort_order"),
        "changeTarget": update_enum_attribute("target", "target"),
        "changeIsActive": update_attribute("isActive", "is_active"),
        "setName": update_attribute("name", "name"),
        "setDescription": update_attribute("description", "description"),
        "setCartPredicate": update_attribute("cartPredicate", "cart_predicate"),
    }



import copy
import typing

from commercetools import types


class Paginator:
    """This paginator uses the offset kwarg for retrieving the pages

    Example::

        paginator = Paginator(client.products.query, sort=["id asc"])
        for product in paginator:
            print(product)

        paginator = Paginator(client.products.query, sort=["id asc"])
        for product in paginator[:-20]:
            print(product)

    """

    def __init__(
        self, operation: typing.Callable, **kwargs: typing.Dict[str, typing.Any]
    ):
        if not callable(operation):
            raise ValueError("Expected a callable as first argument")

        if "offset" in kwargs or "limit" in kwargs:
            raise ValueError(
                "It is not possible to supply either the offset or limit "
                "keyword arguments when using the paginator."
            )

        self.page_size = 20
        self._operation = operation
        self._kwargs = kwargs
        self._offset = 0
        self._limit = None

    def __iter__(self) -> typing.Generator[types.BaseResource, None, None]:

        offset = self._offset or 0
        limit = self._limit
        num = 0
        while True:
            response = self._operation(
                **self._kwargs, offset=offset, limit=self.page_size
            )

            if limit is not None and limit < 0:
                limit = (response.total + limit) - offset

            for item in response.results:
                yield item

                if limit is not None:
                    num += 1
                    if num >= limit:
                        return

            offset += response.count
            if offset >= response.total:
                break

    def _clone(self):
        clone = self.__class__(operation=self._operation, **self._kwargs)
        clone.page_size = self.page_size
        return clone

    def __getitem__(self, item):
        if isinstance(item, slice):
            clone = self._clone()
            clone._offset = item.start
            clone._limit = item.stop
            return clone
        raise IndexError


class CursorPaginator(Paginator):
    """This paginator uses a cursor (where clause) for pagination.

    See https://docs.commercetools.com/http-api.html#paging

    """

    def __iter__(self) -> typing.Generator[types.BaseResource, None, None]:
        last_created_at = None
        limit = self._limit
        num = 0

        kwargs = copy.deepcopy(self._kwargs)
        where_clause = kwargs.setdefault("where", [])
        if where_clause and not isinstance(where_clause, list):
            where_clause = [where_clause]

        kwargs["sort"] = "createdAt asc"

        while True:
            if last_created_at:
                # copy list since we want to append the createdAt filter for
                # every request with an other value.
                kwargs["where"] = list(where_clause)
                kwargs["where"].append(f'createdAt > "{last_created_at.isoformat()}"')

            response = self._operation(**kwargs, limit=self.page_size)
            for item in response.results:
                last_created_at = item.created_at
                yield item
                if limit is not None:
                    num += 1
                    if num >= limit:
                        return

            if response.count < self.page_size:
                break

    def __getitem__(self, item):
        if isinstance(item, slice):
            if item.start:
                raise ValueError("Start slice is not supported")
            clone = self._clone()
            clone._limit = item.stop
            return clone
        raise IndexError



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import (
    BaseResource,
    Reference,
    ReferenceTypeId,
    ResourceIdentifier,
)

if typing.TYPE_CHECKING:
    from ._channel import ChannelResourceIdentifier
    from ._common import CreatedBy, LastModifiedBy
    from ._type import (
        CustomFields,
        CustomFieldsDraft,
        FieldContainer,
        TypeResourceIdentifier,
    )
__all__ = [
    "InventoryEntry",
    "InventoryEntryAddQuantityAction",
    "InventoryEntryChangeQuantityAction",
    "InventoryEntryDraft",
    "InventoryEntryReference",
    "InventoryEntryRemoveQuantityAction",
    "InventoryEntryResourceIdentifier",
    "InventoryEntrySetCustomFieldAction",
    "InventoryEntrySetCustomTypeAction",
    "InventoryEntrySetExpectedDeliveryAction",
    "InventoryEntrySetRestockableInDaysAction",
    "InventoryEntrySetSupplyChannelAction",
    "InventoryEntryUpdate",
    "InventoryEntryUpdateAction",
    "InventoryPagedQueryResponse",
]


class InventoryEntry(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: :class:`str`
    sku: str
    #: Optional :class:`commercetools.types.ChannelResourceIdentifier` `(Named` ``supplyChannel`` `in Commercetools)`
    supply_channel: typing.Optional["ChannelResourceIdentifier"]
    #: :class:`int` `(Named` ``quantityOnStock`` `in Commercetools)`
    quantity_on_stock: int
    #: :class:`int` `(Named` ``availableQuantity`` `in Commercetools)`
    available_quantity: int
    #: Optional :class:`int` `(Named` ``restockableInDays`` `in Commercetools)`
    restockable_in_days: typing.Optional[int]
    #: Optional :class:`datetime.datetime` `(Named` ``expectedDelivery`` `in Commercetools)`
    expected_delivery: typing.Optional[datetime.datetime]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        sku: str,
        quantity_on_stock: int,
        available_quantity: int,
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None,
        supply_channel: typing.Optional["ChannelResourceIdentifier"] = None,
        restockable_in_days: typing.Optional[int] = None,
        expected_delivery: typing.Optional[datetime.datetime] = None,
        custom: typing.Optional["CustomFields"] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.sku = sku
        self.supply_channel = supply_channel
        self.quantity_on_stock = quantity_on_stock
        self.available_quantity = available_quantity
        self.restockable_in_days = restockable_in_days
        self.expected_delivery = expected_delivery
        self.custom = custom
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "InventoryEntry(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, sku=%r, supply_channel=%r, quantity_on_stock=%r, available_quantity=%r, restockable_in_days=%r, expected_delivery=%r, custom=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.sku,
                self.supply_channel,
                self.quantity_on_stock,
                self.available_quantity,
                self.restockable_in_days,
                self.expected_delivery,
                self.custom,
            )
        )


class InventoryEntryDraft(_BaseType):
    #: :class:`str`
    sku: str
    #: Optional :class:`commercetools.types.ChannelResourceIdentifier` `(Named` ``supplyChannel`` `in Commercetools)`
    supply_channel: typing.Optional["ChannelResourceIdentifier"]
    #: :class:`int` `(Named` ``quantityOnStock`` `in Commercetools)`
    quantity_on_stock: int
    #: Optional :class:`int` `(Named` ``restockableInDays`` `in Commercetools)`
    restockable_in_days: typing.Optional[int]
    #: Optional :class:`datetime.datetime` `(Named` ``expectedDelivery`` `in Commercetools)`
    expected_delivery: typing.Optional[datetime.datetime]
    #: Optional :class:`commercetools.types.CustomFieldsDraft`
    custom: typing.Optional["CustomFieldsDraft"]

    def __init__(
        self,
        *,
        sku: str,
        quantity_on_stock: int,
        supply_channel: typing.Optional["ChannelResourceIdentifier"] = None,
        restockable_in_days: typing.Optional[int] = None,
        expected_delivery: typing.Optional[datetime.datetime] = None,
        custom: typing.Optional["CustomFieldsDraft"] = None
    ) -> None:
        self.sku = sku
        self.supply_channel = supply_channel
        self.quantity_on_stock = quantity_on_stock
        self.restockable_in_days = restockable_in_days
        self.expected_delivery = expected_delivery
        self.custom = custom
        super().__init__()

    def __repr__(self) -> str:
        return (
            "InventoryEntryDraft(sku=%r, supply_channel=%r, quantity_on_stock=%r, restockable_in_days=%r, expected_delivery=%r, custom=%r)"
            % (
                self.sku,
                self.supply_channel,
                self.quantity_on_stock,
                self.restockable_in_days,
                self.expected_delivery,
                self.custom,
            )
        )


class InventoryEntryReference(Reference):
    #: Optional :class:`commercetools.types.InventoryEntry`
    obj: typing.Optional["InventoryEntry"]

    def __init__(
        self, *, id: str, obj: typing.Optional["InventoryEntry"] = None
    ) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.INVENTORY_ENTRY, id=id)

    def __repr__(self) -> str:
        return "InventoryEntryReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )


class InventoryEntryResourceIdentifier(ResourceIdentifier):
    def __init__(
        self, *, id: typing.Optional[str] = None, key: typing.Optional[str] = None
    ) -> None:
        super().__init__(type_id=ReferenceTypeId.INVENTORY_ENTRY, id=id, key=key)

    def __repr__(self) -> str:
        return "InventoryEntryResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class InventoryEntryUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "InventoryEntryUpdate(version=%r, actions=%r)" % (
            self.version,
            self.actions,
        )


class InventoryEntryUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "InventoryEntryUpdateAction(action=%r)" % (self.action,)


class InventoryPagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.InventoryEntry`
    results: typing.Sequence["InventoryEntry"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["InventoryEntry"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "InventoryPagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class InventoryEntryAddQuantityAction(InventoryEntryUpdateAction):
    #: :class:`int`
    quantity: int

    def __init__(self, *, quantity: int) -> None:
        self.quantity = quantity
        super().__init__(action="addQuantity")

    def __repr__(self) -> str:
        return "InventoryEntryAddQuantityAction(action=%r, quantity=%r)" % (
            self.action,
            self.quantity,
        )


class InventoryEntryChangeQuantityAction(InventoryEntryUpdateAction):
    #: :class:`int`
    quantity: int

    def __init__(self, *, quantity: int) -> None:
        self.quantity = quantity
        super().__init__(action="changeQuantity")

    def __repr__(self) -> str:
        return "InventoryEntryChangeQuantityAction(action=%r, quantity=%r)" % (
            self.action,
            self.quantity,
        )


class InventoryEntryRemoveQuantityAction(InventoryEntryUpdateAction):
    #: :class:`int`
    quantity: int

    def __init__(self, *, quantity: int) -> None:
        self.quantity = quantity
        super().__init__(action="removeQuantity")

    def __repr__(self) -> str:
        return "InventoryEntryRemoveQuantityAction(action=%r, quantity=%r)" % (
            self.action,
            self.quantity,
        )


class InventoryEntrySetCustomFieldAction(InventoryEntryUpdateAction):
    #: :class:`str`
    name: str
    #: Optional :class:`typing.Any`
    value: typing.Optional[typing.Any]

    def __init__(self, *, name: str, value: typing.Optional[typing.Any] = None) -> None:
        self.name = name
        self.value = value
        super().__init__(action="setCustomField")

    def __repr__(self) -> str:
        return "InventoryEntrySetCustomFieldAction(action=%r, name=%r, value=%r)" % (
            self.action,
            self.name,
            self.value,
        )


class InventoryEntrySetCustomTypeAction(InventoryEntryUpdateAction):
    #: Optional :class:`commercetools.types.TypeResourceIdentifier`
    type: typing.Optional["TypeResourceIdentifier"]
    #: Optional :class:`commercetools.types.FieldContainer`
    fields: typing.Optional["FieldContainer"]

    def __init__(
        self,
        *,
        type: typing.Optional["TypeResourceIdentifier"] = None,
        fields: typing.Optional["FieldContainer"] = None
    ) -> None:
        self.type = type
        self.fields = fields
        super().__init__(action="setCustomType")

    def __repr__(self) -> str:
        return "InventoryEntrySetCustomTypeAction(action=%r, type=%r, fields=%r)" % (
            self.action,
            self.type,
            self.fields,
        )


class InventoryEntrySetExpectedDeliveryAction(InventoryEntryUpdateAction):
    #: Optional :class:`datetime.datetime` `(Named` ``expectedDelivery`` `in Commercetools)`
    expected_delivery: typing.Optional[datetime.datetime]

    def __init__(
        self, *, expected_delivery: typing.Optional[datetime.datetime] = None
    ) -> None:
        self.expected_delivery = expected_delivery
        super().__init__(action="setExpectedDelivery")

    def __repr__(self) -> str:
        return (
            "InventoryEntrySetExpectedDeliveryAction(action=%r, expected_delivery=%r)"
            % (self.action, self.expected_delivery)
        )


class InventoryEntrySetRestockableInDaysAction(InventoryEntryUpdateAction):
    #: Optional :class:`int` `(Named` ``restockableInDays`` `in Commercetools)`
    restockable_in_days: typing.Optional[int]

    def __init__(self, *, restockable_in_days: typing.Optional[int] = None) -> None:
        self.restockable_in_days = restockable_in_days
        super().__init__(action="setRestockableInDays")

    def __repr__(self) -> str:
        return (
            "InventoryEntrySetRestockableInDaysAction(action=%r, restockable_in_days=%r)"
            % (self.action, self.restockable_in_days)
        )


class InventoryEntrySetSupplyChannelAction(InventoryEntryUpdateAction):
    #: Optional :class:`commercetools.types.ChannelResourceIdentifier` `(Named` ``supplyChannel`` `in Commercetools)`
    supply_channel: typing.Optional["ChannelResourceIdentifier"]

    def __init__(
        self, *, supply_channel: typing.Optional["ChannelResourceIdentifier"] = None
    ) -> None:
        self.supply_channel = supply_channel
        super().__init__(action="setSupplyChannel")

    def __repr__(self) -> str:
        return "InventoryEntrySetSupplyChannelAction(action=%r, supply_channel=%r)" % (
            self.action,
            self.supply_channel,
        )



# DO NOT EDIT! This file is automatically generated
import marshmallow
import marshmallow_enum

from commercetools import helpers, types

__all__ = [
    "CartClassificationTypeSchema",
    "CartScoreTypeSchema",
    "CartValueTypeSchema",
    "CartsConfigurationSchema",
    "ExternalOAuthSchema",
    "ProjectChangeCountriesActionSchema",
    "ProjectChangeCountryTaxRateFallbackEnabledActionSchema",
    "ProjectChangeCurrenciesActionSchema",
    "ProjectChangeLanguagesActionSchema",
    "ProjectChangeMessagesConfigurationActionSchema",
    "ProjectChangeMessagesEnabledActionSchema",
    "ProjectChangeNameActionSchema",
    "ProjectSchema",
    "ProjectSetExternalOAuthActionSchema",
    "ProjectSetShippingRateInputTypeActionSchema",
    "ProjectUpdateActionSchema",
    "ProjectUpdateSchema",
    "ShippingRateInputTypeSchema",
]


class CartsConfigurationSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.CartsConfiguration`."""

    country_tax_rate_fallback_enabled = marshmallow.fields.Bool(
        allow_none=True, missing=None, data_key="countryTaxRateFallbackEnabled"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.CartsConfiguration(**data)


class ExternalOAuthSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ExternalOAuth`."""

    url = marshmallow.fields.String(allow_none=True)
    authorization_header = marshmallow.fields.String(
        allow_none=True, data_key="authorizationHeader"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ExternalOAuth(**data)


class ProjectSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.Project`."""

    version = marshmallow.fields.Integer(allow_none=True)
    key = marshmallow.fields.String(allow_none=True)
    name = marshmallow.fields.String(allow_none=True)
    countries = marshmallow.fields.List(marshmallow.fields.String())
    currencies = marshmallow.fields.List(marshmallow.fields.String())
    languages = marshmallow.fields.List(marshmallow.fields.String())
    created_at = marshmallow.fields.DateTime(allow_none=True, data_key="createdAt")
    trial_until = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="trialUntil"
    )
    messages = helpers.LazyNestedField(
        nested="commercetools._schemas._message.MessageConfigurationSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )
    shipping_rate_input_type = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "CartClassification": "commercetools._schemas._project.CartClassificationTypeSchema",
            "CartScore": "commercetools._schemas._project.CartScoreTypeSchema",
            "CartValue": "commercetools._schemas._project.CartValueTypeSchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingRateInputType",
    )
    external_oauth = helpers.LazyNestedField(
        nested="commercetools._schemas._project.ExternalOAuthSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="externalOAuth",
    )
    carts = helpers.LazyNestedField(
        nested="commercetools._schemas._project.CartsConfigurationSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.Project(**data)


class ProjectUpdateActionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProjectUpdateAction`."""

    action = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectUpdateAction(**data)


class ProjectUpdateSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ProjectUpdate`."""

    version = marshmallow.fields.Integer(allow_none=True)
    actions = marshmallow.fields.List(
        helpers.Discriminator(
            discriminator_field=("action", "action"),
            discriminator_schemas={
                "changeCountries": "commercetools._schemas._project.ProjectChangeCountriesActionSchema",
                "changeCountryTaxRateFallbackEnabled": "commercetools._schemas._project.ProjectChangeCountryTaxRateFallbackEnabledActionSchema",
                "changeCurrencies": "commercetools._schemas._project.ProjectChangeCurrenciesActionSchema",
                "changeLanguages": "commercetools._schemas._project.ProjectChangeLanguagesActionSchema",
                "changeMessagesConfiguration": "commercetools._schemas._project.ProjectChangeMessagesConfigurationActionSchema",
                "changeMessagesEnabled": "commercetools._schemas._project.ProjectChangeMessagesEnabledActionSchema",
                "changeName": "commercetools._schemas._project.ProjectChangeNameActionSchema",
                "setExternalOAuth": "commercetools._schemas._project.ProjectSetExternalOAuthActionSchema",
                "setShippingRateInputType": "commercetools._schemas._project.ProjectSetShippingRateInputTypeActionSchema",
            },
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ProjectUpdate(**data)


class ShippingRateInputTypeSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ShippingRateInputType`."""

    type = marshmallow_enum.EnumField(types.ShippingRateTierType, by_value=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.ShippingRateInputType(**data)


class CartClassificationTypeSchema(ShippingRateInputTypeSchema):
    """Marshmallow schema for :class:`commercetools.types.CartClassificationType`."""

    values = marshmallow.fields.List(
        helpers.LazyNestedField(
            nested="commercetools._schemas._type.CustomFieldLocalizedEnumValueSchema",
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CartClassificationType(**data)


class CartScoreTypeSchema(ShippingRateInputTypeSchema):
    """Marshmallow schema for :class:`commercetools.types.CartScoreType`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CartScoreType()


class CartValueTypeSchema(ShippingRateInputTypeSchema):
    """Marshmallow schema for :class:`commercetools.types.CartValueType`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["type"]
        return types.CartValueType()


class ProjectChangeCountriesActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeCountriesAction`."""

    countries = marshmallow.fields.List(marshmallow.fields.String())

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeCountriesAction(**data)


class ProjectChangeCountryTaxRateFallbackEnabledActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeCountryTaxRateFallbackEnabledAction`."""

    country_tax_rate_fallback_enabled = marshmallow.fields.Bool(
        allow_none=True, data_key="countryTaxRateFallbackEnabled"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeCountryTaxRateFallbackEnabledAction(**data)


class ProjectChangeCurrenciesActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeCurrenciesAction`."""

    currencies = marshmallow.fields.List(marshmallow.fields.String())

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeCurrenciesAction(**data)


class ProjectChangeLanguagesActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeLanguagesAction`."""

    languages = marshmallow.fields.List(marshmallow.fields.String())

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeLanguagesAction(**data)


class ProjectChangeMessagesConfigurationActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeMessagesConfigurationAction`."""

    messages_configuration = helpers.LazyNestedField(
        nested="commercetools._schemas._message.MessageConfigurationDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="messagesConfiguration",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeMessagesConfigurationAction(**data)


class ProjectChangeMessagesEnabledActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeMessagesEnabledAction`."""

    messages_enabled = marshmallow.fields.Bool(
        allow_none=True, data_key="messagesEnabled"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeMessagesEnabledAction(**data)


class ProjectChangeNameActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectChangeNameAction`."""

    name = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectChangeNameAction(**data)


class ProjectSetExternalOAuthActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectSetExternalOAuthAction`."""

    external_oauth = helpers.LazyNestedField(
        nested="commercetools._schemas._project.ExternalOAuthSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="externalOAuth",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectSetExternalOAuthAction(**data)


class ProjectSetShippingRateInputTypeActionSchema(ProjectUpdateActionSchema):
    """Marshmallow schema for :class:`commercetools.types.ProjectSetShippingRateInputTypeAction`."""

    shipping_rate_input_type = helpers.Discriminator(
        discriminator_field=("type", "type"),
        discriminator_schemas={
            "CartClassification": "commercetools._schemas._project.CartClassificationTypeSchema",
            "CartScore": "commercetools._schemas._project.CartScoreTypeSchema",
            "CartValue": "commercetools._schemas._project.CartValueTypeSchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="shippingRateInputType",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["action"]
        return types.ProjectSetShippingRateInputTypeAction(**data)



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._customer import (
    CustomerChangePasswordSchema,
    CustomerCreateEmailTokenSchema,
    CustomerCreatePasswordResetTokenSchema,
    CustomerDraftSchema,
    CustomerEmailVerifySchema,
    CustomerPagedQueryResponseSchema,
    CustomerResetPasswordSchema,
    CustomerSchema,
    CustomerSignInResultSchema,
    CustomerTokenSchema,
    CustomerUpdateSchema,
)
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._customer import (
    Customer,
    CustomerChangePassword,
    CustomerCreateEmailToken,
    CustomerCreatePasswordResetToken,
    CustomerDraft,
    CustomerEmailVerify,
    CustomerPagedQueryResponse,
    CustomerResetPassword,
    CustomerSignInResult,
    CustomerToken,
    CustomerUpdate,
    CustomerUpdateAction,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _CustomerQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
):
    pass


class _CustomerUpdateSchema(traits.ExpandableSchema, traits.VersionedSchema):
    pass


class _CustomerDeleteSchema(
    traits.VersionedSchema, traits.ExpandableSchema, traits.DataErasureSchema
):
    pass


class CustomerService(abstract.AbstractService):
    """A customer is a person purchasing products.

    customers, Orders, Comments and Reviews can be associated to a customer.
    """

    def get_by_email_token(
        self, email_token: str, *, expand: OptionalListStr = None
    ) -> Customer:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"customers/email-token={email_token}",
            params=params,
            schema_cls=CustomerSchema,
        )

    def get_by_id(self, id: str, *, expand: OptionalListStr = None) -> Customer:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"customers/{id}", params=params, schema_cls=CustomerSchema
        )

    def get_by_key(self, key: str, *, expand: OptionalListStr = None) -> Customer:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"customers/key={key}", params=params, schema_cls=CustomerSchema
        )

    def get_by_password_token(
        self, password_token: str, *, expand: OptionalListStr = None
    ) -> Customer:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"customers/password-token={password_token}",
            params=params,
            schema_cls=CustomerSchema,
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> CustomerPagedQueryResponse:
        """A customer is a person purchasing products. customers, Orders, Comments
        and Reviews can be associated to a customer.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _CustomerQuerySchema,
        )
        return self._client._get(
            endpoint="customers",
            params=params,
            schema_cls=CustomerPagedQueryResponseSchema,
        )

    def create(
        self, draft: CustomerDraft, *, expand: OptionalListStr = None
    ) -> CustomerSignInResult:
        """Creates a customer.

        If an anonymous cart is passed in, then the cart is assigned to the
        created customer and the version number of the Cart will increase. If the
        ID of an anonymous session is given, all carts and orders will be
        assigned to the created customer.   A customer is a person purchasing
        products. customers, Orders, Comments and Reviews can be associated to a
        customer.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="customers",
            params=params,
            data_object=draft,
            request_schema_cls=CustomerDraftSchema,
            response_schema_cls=CustomerSignInResultSchema,
        )

    def update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[CustomerUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> Customer:
        params = self._serialize_params({"expand": expand}, _CustomerUpdateSchema)
        update_action = CustomerUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"customers/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=CustomerUpdateSchema,
            response_schema_cls=CustomerSchema,
            force_update=force_update,
        )

    def update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[CustomerUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> Customer:
        params = self._serialize_params({"expand": expand}, _CustomerUpdateSchema)
        update_action = CustomerUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"customers/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=CustomerUpdateSchema,
            response_schema_cls=CustomerSchema,
            force_update=force_update,
        )

    def delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        data_erasure: bool = None,
        force_delete: bool = False,
    ) -> Customer:
        params = self._serialize_params(
            {"version": version, "expand": expand, "data_erasure": data_erasure},
            _CustomerDeleteSchema,
        )
        return self._client._delete(
            endpoint=f"customers/{id}",
            params=params,
            response_schema_cls=CustomerSchema,
            force_delete=force_delete,
        )

    def delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        data_erasure: bool = None,
        force_delete: bool = False,
    ) -> Customer:
        params = self._serialize_params(
            {"version": version, "expand": expand, "data_erasure": data_erasure},
            _CustomerDeleteSchema,
        )
        return self._client._delete(
            endpoint=f"customers/key={key}",
            params=params,
            response_schema_cls=CustomerSchema,
            force_delete=force_delete,
        )

    def email_confirm(self, action: CustomerEmailVerify) -> Customer:
        """Verifies customer's email using a token."""
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="customers/email/confirm",
            params=params,
            data_object=action,
            request_schema_cls=CustomerEmailVerifySchema,
            response_schema_cls=CustomerSchema,
        )

    def email_token(self, action: CustomerCreateEmailToken) -> CustomerToken:
        """Create a Token for verifying the Customer's Email

        To verify a customer's email, an email token can be created. This should
        be embedded in a link and sent to the customer via email. When the
        customer clicks on the link, the "verify customer's email" endpoint
        should be called, which sets customer's isVerifiedEmail field to true.
        """
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="customers/email-token",
            params=params,
            data_object=action,
            request_schema_cls=CustomerCreateEmailTokenSchema,
            response_schema_cls=CustomerTokenSchema,
        )

    def password(self, action: CustomerChangePassword) -> Customer:
        """Change a customers password"""
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="customers/password",
            params=params,
            data_object=action,
            request_schema_cls=CustomerChangePasswordSchema,
            response_schema_cls=CustomerSchema,
        )

    def password_reset(self, action: CustomerResetPassword) -> Customer:
        """Set a new password using a token."""
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="customers/password/reset",
            params=params,
            data_object=action,
            request_schema_cls=CustomerResetPasswordSchema,
            response_schema_cls=CustomerSchema,
        )

    def password_token(self, action: CustomerCreatePasswordResetToken) -> CustomerToken:
        """The token value is used to reset the password of the customer with the
        given email.

        The token is valid only for 10 minutes.   The following workflow can be
        used to reset the customer's password:  * Create a password reset token
        and send it embedded in a link to the customer. * When the customer
        clicks on the link, the customer is retrieved with the token. * The
        customer enters a new password and the "reset customer's password"
        endpoint is called.
        """
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="customers/password-token",
            params=params,
            data_object=action,
            request_schema_cls=CustomerCreatePasswordResetTokenSchema,
            response_schema_cls=CustomerTokenSchema,
        )



from commercetools import types


def test_extension_create(client):
    extension = client.extensions.create(
        types.ExtensionDraft(
            destination=types.ExtensionAWSLambdaDestination(
                arn="arn:", access_key="access", access_secret="secret"
            ),
            triggers=[],
        )
    )
    assert extension.id


def test_extension_get_by_id(client):
    extension = client.extensions.create(
        types.ExtensionDraft(
            destination=types.ExtensionAWSLambdaDestination(
                arn="arn:", access_key="access", access_secret="secret"
            ),
            triggers=[],
        )
    )
    assert extension.id

    extension = client.extensions.get_by_id(extension.id)
    assert extension.id



import copy
import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._inventory import (
    InventoryEntryDraftSchema,
    InventoryEntrySchema,
    InventoryEntryUpdateSchema,
    InventoryPagedQueryResponseSchema,
)
from commercetools.testing import utils
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import InternalUpdateError, update_attribute


class InventoryEntryModel(BaseModel):
    _primary_type_name = "inventory-entry"
    _resource_schema = InventoryEntrySchema

    def _create_from_draft(
        self, draft: types.InventoryEntryDraft, id: typing.Optional[str] = None
    ) -> types.InventoryEntry:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.InventoryEntry(
            id=str(object_id),
            version=1,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            expected_delivery=draft.expected_delivery,
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            quantity_on_stock=draft.quantity_on_stock,
            available_quantity=draft.quantity_on_stock,
            restockable_in_days=draft.restockable_in_days,
            sku=draft.sku,
            supply_channel=draft.supply_channel,
            custom=utils.create_from_draft(draft.custom),
        )


def change_stock():
    def updater(self, obj, action: types.InventoryEntryUpdateAction):
        quantity = getattr(action, "quantity")

        new = copy.deepcopy(obj)

        if isinstance(action, types.InventoryEntryAddQuantityAction):
            new["availableQuantity"] += quantity
        elif isinstance(action, types.InventoryEntryRemoveQuantityAction):
            new["availableQuantity"] -= quantity
        elif isinstance(action, types.InventoryEntryChangeQuantityAction):
            new["availableQuantity"] = quantity
        else:
            raise InternalUpdateError("Unknown action to change stock: %r", action)

        # For now we don't reserve stock
        new["quantityOnStock"] = new["availableQuantity"]
        return new

    return updater


class InventoryEntryBackend(ServiceBackend):
    service_path = "inventory"
    model_class = InventoryEntryModel
    _schema_draft = InventoryEntryDraftSchema
    _schema_update = InventoryEntryUpdateSchema
    _schema_query_response = InventoryPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
        ]

    _actions = {
        "addQuantity": change_stock(),
        "removeQuantity": change_stock(),
        "changeQuantity": change_stock(),
        "setRestockableInDays": update_attribute(
            "restockableInDays", "restockableInDays"
        ),
        "setExpectedDelivery": update_attribute("expectedDelivery", "expectedDelivery"),
    }



# DO NOT EDIT! This file is automatically generated
import datetime
import enum
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import (
    BaseResource,
    Reference,
    ReferenceTypeId,
    ResourceIdentifier,
)

if typing.TYPE_CHECKING:
    from ._channel import ChannelReference
    from ._common import CreatedBy, LastModifiedBy, LocalizedString, Money, TypedMoney
    from ._product import ProductReference
    from ._type import CustomFields, TypeResourceIdentifier
__all__ = [
    "CartDiscount",
    "CartDiscountChangeCartPredicateAction",
    "CartDiscountChangeIsActiveAction",
    "CartDiscountChangeNameAction",
    "CartDiscountChangeRequiresDiscountCodeAction",
    "CartDiscountChangeSortOrderAction",
    "CartDiscountChangeStackingModeAction",
    "CartDiscountChangeTargetAction",
    "CartDiscountChangeValueAction",
    "CartDiscountCustomLineItemsTarget",
    "CartDiscountDraft",
    "CartDiscountLineItemsTarget",
    "CartDiscountPagedQueryResponse",
    "CartDiscountReference",
    "CartDiscountResourceIdentifier",
    "CartDiscountSetCustomFieldAction",
    "CartDiscountSetCustomTypeAction",
    "CartDiscountSetDescriptionAction",
    "CartDiscountSetKeyAction",
    "CartDiscountSetValidFromAction",
    "CartDiscountSetValidFromAndUntilAction",
    "CartDiscountSetValidUntilAction",
    "CartDiscountShippingCostTarget",
    "CartDiscountTarget",
    "CartDiscountUpdate",
    "CartDiscountUpdateAction",
    "CartDiscountValue",
    "CartDiscountValueAbsolute",
    "CartDiscountValueAbsoluteDraft",
    "CartDiscountValueDraft",
    "CartDiscountValueGiftLineItem",
    "CartDiscountValueGiftLineItemDraft",
    "CartDiscountValueRelative",
    "CartDiscountValueRelativeDraft",
    "MultiBuyCustomLineItemsTarget",
    "MultiBuyLineItemsTarget",
    "SelectionMode",
    "StackingMode",
]


class CartDiscount(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: :class:`commercetools.types.LocalizedString`
    name: "LocalizedString"
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]
    #: :class:`commercetools.types.CartDiscountValue`
    value: "CartDiscountValue"
    #: :class:`str` `(Named` ``cartPredicate`` `in Commercetools)`
    cart_predicate: str
    #: Optional :class:`commercetools.types.CartDiscountTarget`
    target: typing.Optional["CartDiscountTarget"]
    #: :class:`str` `(Named` ``sortOrder`` `in Commercetools)`
    sort_order: str
    #: :class:`bool` `(Named` ``isActive`` `in Commercetools)`
    is_active: bool
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: :class:`bool` `(Named` ``requiresDiscountCode`` `in Commercetools)`
    requires_discount_code: bool
    #: List of :class:`commercetools.types.Reference`
    references: typing.List["Reference"]
    #: :class:`commercetools.types.StackingMode` `(Named` ``stackingMode`` `in Commercetools)`
    stacking_mode: "StackingMode"
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        name: "LocalizedString",
        value: "CartDiscountValue",
        cart_predicate: str,
        sort_order: str,
        is_active: bool,
        requires_discount_code: bool,
        references: typing.List["Reference"],
        stacking_mode: "StackingMode",
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None,
        key: typing.Optional[str] = None,
        description: typing.Optional["LocalizedString"] = None,
        target: typing.Optional["CartDiscountTarget"] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        custom: typing.Optional["CustomFields"] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.name = name
        self.key = key
        self.description = description
        self.value = value
        self.cart_predicate = cart_predicate
        self.target = target
        self.sort_order = sort_order
        self.is_active = is_active
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.requires_discount_code = requires_discount_code
        self.references = references
        self.stacking_mode = stacking_mode
        self.custom = custom
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "CartDiscount(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, name=%r, key=%r, description=%r, value=%r, cart_predicate=%r, target=%r, sort_order=%r, is_active=%r, valid_from=%r, valid_until=%r, requires_discount_code=%r, references=%r, stacking_mode=%r, custom=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.name,
                self.key,
                self.description,
                self.value,
                self.cart_predicate,
                self.target,
                self.sort_order,
                self.is_active,
                self.valid_from,
                self.valid_until,
                self.requires_discount_code,
                self.references,
                self.stacking_mode,
                self.custom,
            )
        )


class CartDiscountDraft(_BaseType):
    #: :class:`commercetools.types.LocalizedString`
    name: "LocalizedString"
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]
    #: :class:`commercetools.types.CartDiscountValueDraft`
    value: "CartDiscountValueDraft"
    #: :class:`str` `(Named` ``cartPredicate`` `in Commercetools)`
    cart_predicate: str
    #: Optional :class:`commercetools.types.CartDiscountTarget`
    target: typing.Optional["CartDiscountTarget"]
    #: :class:`str` `(Named` ``sortOrder`` `in Commercetools)`
    sort_order: str
    #: Optional :class:`bool` `(Named` ``isActive`` `in Commercetools)`
    is_active: typing.Optional[bool]
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: :class:`bool` `(Named` ``requiresDiscountCode`` `in Commercetools)`
    requires_discount_code: bool
    #: Optional :class:`commercetools.types.StackingMode` `(Named` ``stackingMode`` `in Commercetools)`
    stacking_mode: typing.Optional["StackingMode"]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]

    def __init__(
        self,
        *,
        name: "LocalizedString",
        value: "CartDiscountValueDraft",
        cart_predicate: str,
        sort_order: str,
        requires_discount_code: bool,
        key: typing.Optional[str] = None,
        description: typing.Optional["LocalizedString"] = None,
        target: typing.Optional["CartDiscountTarget"] = None,
        is_active: typing.Optional[bool] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        stacking_mode: typing.Optional["StackingMode"] = None,
        custom: typing.Optional["CustomFields"] = None
    ) -> None:
        self.name = name
        self.key = key
        self.description = description
        self.value = value
        self.cart_predicate = cart_predicate
        self.target = target
        self.sort_order = sort_order
        self.is_active = is_active
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.requires_discount_code = requires_discount_code
        self.stacking_mode = stacking_mode
        self.custom = custom
        super().__init__()

    def __repr__(self) -> str:
        return (
            "CartDiscountDraft(name=%r, key=%r, description=%r, value=%r, cart_predicate=%r, target=%r, sort_order=%r, is_active=%r, valid_from=%r, valid_until=%r, requires_discount_code=%r, stacking_mode=%r, custom=%r)"
            % (
                self.name,
                self.key,
                self.description,
                self.value,
                self.cart_predicate,
                self.target,
                self.sort_order,
                self.is_active,
                self.valid_from,
                self.valid_until,
                self.requires_discount_code,
                self.stacking_mode,
                self.custom,
            )
        )


class CartDiscountPagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.CartDiscount`
    results: typing.Sequence["CartDiscount"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["CartDiscount"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "CartDiscountPagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class CartDiscountReference(Reference):
    #: Optional :class:`commercetools.types.CartDiscount`
    obj: typing.Optional["CartDiscount"]

    def __init__(self, *, id: str, obj: typing.Optional["CartDiscount"] = None) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.CART_DISCOUNT, id=id)

    def __repr__(self) -> str:
        return "CartDiscountReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )


class CartDiscountResourceIdentifier(ResourceIdentifier):
    def __init__(
        self, *, id: typing.Optional[str] = None, key: typing.Optional[str] = None
    ) -> None:
        super().__init__(type_id=ReferenceTypeId.CART_DISCOUNT, id=id, key=key)

    def __repr__(self) -> str:
        return "CartDiscountResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class CartDiscountTarget(_BaseType):
    #: :class:`str`
    type: str

    def __init__(self, *, type: str) -> None:
        self.type = type
        super().__init__()

    def __repr__(self) -> str:
        return "CartDiscountTarget(type=%r)" % (self.type,)


class CartDiscountUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "CartDiscountUpdate(version=%r, actions=%r)" % (
            self.version,
            self.actions,
        )


class CartDiscountUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "CartDiscountUpdateAction(action=%r)" % (self.action,)


class CartDiscountValue(_BaseType):
    #: :class:`str`
    type: str

    def __init__(self, *, type: str) -> None:
        self.type = type
        super().__init__()

    def __repr__(self) -> str:
        return "CartDiscountValue(type=%r)" % (self.type,)


class CartDiscountValueDraft(_BaseType):
    #: :class:`str`
    type: str

    def __init__(self, *, type: str) -> None:
        self.type = type
        super().__init__()

    def __repr__(self) -> str:
        return "CartDiscountValueDraft(type=%r)" % (self.type,)


class SelectionMode(enum.Enum):
    CHEAPEST = "Cheapest"
    MOST_EXPENSIVE = "MostExpensive"


class StackingMode(enum.Enum):
    STACKING = "Stacking"
    STOP_AFTER_THIS_DISCOUNT = "StopAfterThisDiscount"


class CartDiscountChangeCartPredicateAction(CartDiscountUpdateAction):
    #: :class:`str` `(Named` ``cartPredicate`` `in Commercetools)`
    cart_predicate: str

    def __init__(self, *, cart_predicate: str) -> None:
        self.cart_predicate = cart_predicate
        super().__init__(action="changeCartPredicate")

    def __repr__(self) -> str:
        return "CartDiscountChangeCartPredicateAction(action=%r, cart_predicate=%r)" % (
            self.action,
            self.cart_predicate,
        )


class CartDiscountChangeIsActiveAction(CartDiscountUpdateAction):
    #: :class:`bool` `(Named` ``isActive`` `in Commercetools)`
    is_active: bool

    def __init__(self, *, is_active: bool) -> None:
        self.is_active = is_active
        super().__init__(action="changeIsActive")

    def __repr__(self) -> str:
        return "CartDiscountChangeIsActiveAction(action=%r, is_active=%r)" % (
            self.action,
            self.is_active,
        )


class CartDiscountChangeNameAction(CartDiscountUpdateAction):
    #: :class:`commercetools.types.LocalizedString`
    name: "LocalizedString"

    def __init__(self, *, name: "LocalizedString") -> None:
        self.name = name
        super().__init__(action="changeName")

    def __repr__(self) -> str:
        return "CartDiscountChangeNameAction(action=%r, name=%r)" % (
            self.action,
            self.name,
        )


class CartDiscountChangeRequiresDiscountCodeAction(CartDiscountUpdateAction):
    #: :class:`bool` `(Named` ``requiresDiscountCode`` `in Commercetools)`
    requires_discount_code: bool

    def __init__(self, *, requires_discount_code: bool) -> None:
        self.requires_discount_code = requires_discount_code
        super().__init__(action="changeRequiresDiscountCode")

    def __repr__(self) -> str:
        return (
            "CartDiscountChangeRequiresDiscountCodeAction(action=%r, requires_discount_code=%r)"
            % (self.action, self.requires_discount_code)
        )


class CartDiscountChangeSortOrderAction(CartDiscountUpdateAction):
    #: :class:`str` `(Named` ``sortOrder`` `in Commercetools)`
    sort_order: str

    def __init__(self, *, sort_order: str) -> None:
        self.sort_order = sort_order
        super().__init__(action="changeSortOrder")

    def __repr__(self) -> str:
        return "CartDiscountChangeSortOrderAction(action=%r, sort_order=%r)" % (
            self.action,
            self.sort_order,
        )


class CartDiscountChangeStackingModeAction(CartDiscountUpdateAction):
    #: :class:`commercetools.types.StackingMode` `(Named` ``stackingMode`` `in Commercetools)`
    stacking_mode: "StackingMode"

    def __init__(self, *, stacking_mode: "StackingMode") -> None:
        self.stacking_mode = stacking_mode
        super().__init__(action="changeStackingMode")

    def __repr__(self) -> str:
        return "CartDiscountChangeStackingModeAction(action=%r, stacking_mode=%r)" % (
            self.action,
            self.stacking_mode,
        )


class CartDiscountChangeTargetAction(CartDiscountUpdateAction):
    #: :class:`commercetools.types.CartDiscountTarget`
    target: "CartDiscountTarget"

    def __init__(self, *, target: "CartDiscountTarget") -> None:
        self.target = target
        super().__init__(action="changeTarget")

    def __repr__(self) -> str:
        return "CartDiscountChangeTargetAction(action=%r, target=%r)" % (
            self.action,
            self.target,
        )


class CartDiscountChangeValueAction(CartDiscountUpdateAction):
    #: :class:`commercetools.types.CartDiscountValueDraft`
    value: "CartDiscountValueDraft"

    def __init__(self, *, value: "CartDiscountValueDraft") -> None:
        self.value = value
        super().__init__(action="changeValue")

    def __repr__(self) -> str:
        return "CartDiscountChangeValueAction(action=%r, value=%r)" % (
            self.action,
            self.value,
        )


class CartDiscountCustomLineItemsTarget(CartDiscountTarget):
    #: :class:`str`
    predicate: str

    def __init__(self, *, predicate: str) -> None:
        self.predicate = predicate
        super().__init__(type="customLineItems")

    def __repr__(self) -> str:
        return "CartDiscountCustomLineItemsTarget(type=%r, predicate=%r)" % (
            self.type,
            self.predicate,
        )


class CartDiscountLineItemsTarget(CartDiscountTarget):
    #: :class:`str`
    predicate: str

    def __init__(self, *, predicate: str) -> None:
        self.predicate = predicate
        super().__init__(type="lineItems")

    def __repr__(self) -> str:
        return "CartDiscountLineItemsTarget(type=%r, predicate=%r)" % (
            self.type,
            self.predicate,
        )


class CartDiscountSetCustomFieldAction(CartDiscountUpdateAction):
    #: :class:`str`
    name: str
    #: Optional :class:`typing.Any`
    value: typing.Optional[typing.Any]

    def __init__(self, *, name: str, value: typing.Optional[typing.Any] = None) -> None:
        self.name = name
        self.value = value
        super().__init__(action="setCustomField")

    def __repr__(self) -> str:
        return "CartDiscountSetCustomFieldAction(action=%r, name=%r, value=%r)" % (
            self.action,
            self.name,
            self.value,
        )


class CartDiscountSetCustomTypeAction(CartDiscountUpdateAction):
    #: Optional :class:`commercetools.types.TypeResourceIdentifier`
    type: typing.Optional["TypeResourceIdentifier"]
    #: Optional :class:`object`
    fields: typing.Optional[object]

    def __init__(
        self,
        *,
        type: typing.Optional["TypeResourceIdentifier"] = None,
        fields: typing.Optional[object] = None
    ) -> None:
        self.type = type
        self.fields = fields
        super().__init__(action="setCustomType")

    def __repr__(self) -> str:
        return "CartDiscountSetCustomTypeAction(action=%r, type=%r, fields=%r)" % (
            self.action,
            self.type,
            self.fields,
        )


class CartDiscountSetDescriptionAction(CartDiscountUpdateAction):
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]

    def __init__(
        self, *, description: typing.Optional["LocalizedString"] = None
    ) -> None:
        self.description = description
        super().__init__(action="setDescription")

    def __repr__(self) -> str:
        return "CartDiscountSetDescriptionAction(action=%r, description=%r)" % (
            self.action,
            self.description,
        )


class CartDiscountSetKeyAction(CartDiscountUpdateAction):
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(self, *, key: typing.Optional[str] = None) -> None:
        self.key = key
        super().__init__(action="setKey")

    def __repr__(self) -> str:
        return "CartDiscountSetKeyAction(action=%r, key=%r)" % (self.action, self.key)


class CartDiscountSetValidFromAction(CartDiscountUpdateAction):
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]

    def __init__(
        self, *, valid_from: typing.Optional[datetime.datetime] = None
    ) -> None:
        self.valid_from = valid_from
        super().__init__(action="setValidFrom")

    def __repr__(self) -> str:
        return "CartDiscountSetValidFromAction(action=%r, valid_from=%r)" % (
            self.action,
            self.valid_from,
        )


class CartDiscountSetValidFromAndUntilAction(CartDiscountUpdateAction):
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]

    def __init__(
        self,
        *,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None
    ) -> None:
        self.valid_from = valid_from
        self.valid_until = valid_until
        super().__init__(action="setValidFromAndUntil")

    def __repr__(self) -> str:
        return (
            "CartDiscountSetValidFromAndUntilAction(action=%r, valid_from=%r, valid_until=%r)"
            % (self.action, self.valid_from, self.valid_until)
        )


class CartDiscountSetValidUntilAction(CartDiscountUpdateAction):
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]

    def __init__(
        self, *, valid_until: typing.Optional[datetime.datetime] = None
    ) -> None:
        self.valid_until = valid_until
        super().__init__(action="setValidUntil")

    def __repr__(self) -> str:
        return "CartDiscountSetValidUntilAction(action=%r, valid_until=%r)" % (
            self.action,
            self.valid_until,
        )


class CartDiscountShippingCostTarget(CartDiscountTarget):
    def __init__(self) -> None:
        super().__init__(type="shipping")

    def __repr__(self) -> str:
        return "CartDiscountShippingCostTarget(type=%r)" % (self.type,)


class CartDiscountValueAbsolute(CartDiscountValue):
    #: List of :class:`commercetools.types.TypedMoney`
    money: typing.List["TypedMoney"]

    def __init__(self, *, money: typing.List["TypedMoney"]) -> None:
        self.money = money
        super().__init__(type="absolute")

    def __repr__(self) -> str:
        return "CartDiscountValueAbsolute(type=%r, money=%r)" % (self.type, self.money)


class CartDiscountValueAbsoluteDraft(CartDiscountValueDraft):
    #: List of :class:`commercetools.types.Money`
    money: typing.List["Money"]

    def __init__(self, *, money: typing.List["Money"]) -> None:
        self.money = money
        super().__init__(type="absolute")

    def __repr__(self) -> str:
        return "CartDiscountValueAbsoluteDraft(type=%r, money=%r)" % (
            self.type,
            self.money,
        )


class CartDiscountValueGiftLineItem(CartDiscountValue):
    #: :class:`commercetools.types.ProductReference`
    product: "ProductReference"
    #: :class:`int` `(Named` ``variantId`` `in Commercetools)`
    variant_id: int
    #: Optional :class:`commercetools.types.ChannelReference` `(Named` ``supplyChannel`` `in Commercetools)`
    supply_channel: typing.Optional["ChannelReference"]
    #: Optional :class:`commercetools.types.ChannelReference` `(Named` ``distributionChannel`` `in Commercetools)`
    distribution_channel: typing.Optional["ChannelReference"]

    def __init__(
        self,
        *,
        product: "ProductReference",
        variant_id: int,
        supply_channel: typing.Optional["ChannelReference"] = None,
        distribution_channel: typing.Optional["ChannelReference"] = None
    ) -> None:
        self.product = product
        self.variant_id = variant_id
        self.supply_channel = supply_channel
        self.distribution_channel = distribution_channel
        super().__init__(type="giftLineItem")

    def __repr__(self) -> str:
        return (
            "CartDiscountValueGiftLineItem(type=%r, product=%r, variant_id=%r, supply_channel=%r, distribution_channel=%r)"
            % (
                self.type,
                self.product,
                self.variant_id,
                self.supply_channel,
                self.distribution_channel,
            )
        )


class CartDiscountValueGiftLineItemDraft(CartDiscountValueDraft):
    #: :class:`commercetools.types.ProductReference`
    product: "ProductReference"
    #: :class:`int` `(Named` ``variantId`` `in Commercetools)`
    variant_id: int
    #: Optional :class:`commercetools.types.ChannelReference` `(Named` ``supplyChannel`` `in Commercetools)`
    supply_channel: typing.Optional["ChannelReference"]
    #: Optional :class:`commercetools.types.ChannelReference` `(Named` ``distributionChannel`` `in Commercetools)`
    distribution_channel: typing.Optional["ChannelReference"]

    def __init__(
        self,
        *,
        product: "ProductReference",
        variant_id: int,
        supply_channel: typing.Optional["ChannelReference"] = None,
        distribution_channel: typing.Optional["ChannelReference"] = None
    ) -> None:
        self.product = product
        self.variant_id = variant_id
        self.supply_channel = supply_channel
        self.distribution_channel = distribution_channel
        super().__init__(type="giftLineItem")

    def __repr__(self) -> str:
        return (
            "CartDiscountValueGiftLineItemDraft(type=%r, product=%r, variant_id=%r, supply_channel=%r, distribution_channel=%r)"
            % (
                self.type,
                self.product,
                self.variant_id,
                self.supply_channel,
                self.distribution_channel,
            )
        )


class CartDiscountValueRelative(CartDiscountValue):
    #: :class:`int`
    permyriad: int

    def __init__(self, *, permyriad: int) -> None:
        self.permyriad = permyriad
        super().__init__(type="relative")

    def __repr__(self) -> str:
        return "CartDiscountValueRelative(type=%r, permyriad=%r)" % (
            self.type,
            self.permyriad,
        )


class CartDiscountValueRelativeDraft(CartDiscountValueDraft):
    #: :class:`int`
    permyriad: int

    def __init__(self, *, permyriad: int) -> None:
        self.permyriad = permyriad
        super().__init__(type="relative")

    def __repr__(self) -> str:
        return "CartDiscountValueRelativeDraft(type=%r, permyriad=%r)" % (
            self.type,
            self.permyriad,
        )


class MultiBuyCustomLineItemsTarget(CartDiscountTarget):
    #: :class:`str`
    predicate: str
    #: :class:`int` `(Named` ``triggerQuantity`` `in Commercetools)`
    trigger_quantity: int
    #: :class:`int` `(Named` ``discountedQuantity`` `in Commercetools)`
    discounted_quantity: int
    #: Optional :class:`int` `(Named` ``maxOccurrence`` `in Commercetools)`
    max_occurrence: typing.Optional[int]
    #: :class:`commercetools.types.SelectionMode` `(Named` ``selectionMode`` `in Commercetools)`
    selection_mode: "SelectionMode"

    def __init__(
        self,
        *,
        predicate: str,
        trigger_quantity: int,
        discounted_quantity: int,
        selection_mode: "SelectionMode",
        max_occurrence: typing.Optional[int] = None
    ) -> None:
        self.predicate = predicate
        self.trigger_quantity = trigger_quantity
        self.discounted_quantity = discounted_quantity
        self.max_occurrence = max_occurrence
        self.selection_mode = selection_mode
        super().__init__(type="multiBuyCustomLineItems")

    def __repr__(self) -> str:
        return (
            "MultiBuyCustomLineItemsTarget(type=%r, predicate=%r, trigger_quantity=%r, discounted_quantity=%r, max_occurrence=%r, selection_mode=%r)"
            % (
                self.type,
                self.predicate,
                self.trigger_quantity,
                self.discounted_quantity,
                self.max_occurrence,
                self.selection_mode,
            )
        )


class MultiBuyLineItemsTarget(CartDiscountTarget):
    #: :class:`str`
    predicate: str
    #: :class:`int` `(Named` ``triggerQuantity`` `in Commercetools)`
    trigger_quantity: int
    #: :class:`int` `(Named` ``discountedQuantity`` `in Commercetools)`
    discounted_quantity: int
    #: Optional :class:`int` `(Named` ``maxOccurrence`` `in Commercetools)`
    max_occurrence: typing.Optional[int]
    #: :class:`commercetools.types.SelectionMode` `(Named` ``selectionMode`` `in Commercetools)`
    selection_mode: "SelectionMode"

    def __init__(
        self,
        *,
        predicate: str,
        trigger_quantity: int,
        discounted_quantity: int,
        selection_mode: "SelectionMode",
        max_occurrence: typing.Optional[int] = None
    ) -> None:
        self.predicate = predicate
        self.trigger_quantity = trigger_quantity
        self.discounted_quantity = discounted_quantity
        self.max_occurrence = max_occurrence
        self.selection_mode = selection_mode
        super().__init__(type="multiBuyLineItems")

    def __repr__(self) -> str:
        return (
            "MultiBuyLineItemsTarget(type=%r, predicate=%r, trigger_quantity=%r, discounted_quantity=%r, max_occurrence=%r, selection_mode=%r)"
            % (
                self.type,
                self.predicate,
                self.trigger_quantity,
                self.discounted_quantity,
                self.max_occurrence,
                self.selection_mode,
            )
        )



import pytest

from commercetools._schemas._product import ProductDataSchema
from commercetools.testing import predicates


@pytest.mark.parametrize(
    "predicate",
    [
        "age <> 42",
        "age >= 42",
        "age <= 42",
        r'not (name = "P\"eter" and age < 42)',
        "age not in (42, 43, 44)",
        'tags contains all ("a", "b", "c")',
        'tags contains all ("a", "b", "c")',
        'name = "Peter" or name = "Barbara D"',
        'name in ("Peter", "Barbara")',
        "name is not defined",
        "geoLocation within circle(13.37770, 52.51627, 1000)",
        "isDefault = true",
        "isDefault = false",
        'variants(attributes(name="attribute-name" and value(centAmount > 999 and centAmount < 1001 and currencyCode="EUR")))',
        'custom(fields(someCustomField="123123123")) and createdAt >= "2019-10-15T14:12:36.464465"',
    ],
)
def test_tokenize(predicate):
    predicates.PredicateFilter(predicate, schema=None)


@pytest.mark.parametrize(
    "predicate,paths",
    [
        (
            'slug(nl-uk = "test-categorie" or nl-nl = "test-categorie")',
            [["slug", "nl-uk"], ["slug", "nl-nl"]],
        ),
        (
            'slug(nl-be = "test-categorie") and masterVariant is not defined',
            [["slug", "nl-be"], ["masterVariant"]],
        ),
    ],
)
def test_filter(predicate, paths):
    pf = predicates.PredicateFilter(predicate, schema=ProductDataSchema)

    found_paths = []

    org_filter_field = pf.filter_field

    def mock_filter_field(obj, path, operator, value):
        found_paths.append(path)
        return org_filter_field(obj, path, operator, value)

    pf.filter_field = mock_filter_field
    pf.match({"slug": {"nl-BE": "test-categorie"}})
    assert paths == found_paths


def test_in_filter():
    skus = '("1337", "1338")'
    predicate = f"masterVariant(sku in {skus}) or variants(sku in {skus})"
    pf = predicates.PredicateFilter(predicate, schema=ProductDataSchema)

    product_data = {
        "masterVariant": {"sku": "1337"},
        "variants": [{"sku": "1338"}, {"sku": "1339"}],
    }
    assert pf.match(product_data) is True

    product_data = {
        "masterVariant": {"sku": "1444"},
        "variants": [{"sku": "1338"}, {"sku": "1339"}],
    }
    assert pf.match(product_data) is True

    product_data = {
        "masterVariant": {"sku": "1444"},
        "variants": [{"sku": "1438"}, {"sku": "1139"}],
    }
    assert pf.match(product_data) is False



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._store import (
    StoreDraftSchema,
    StorePagedQueryResponseSchema,
    StoreSchema,
    StoreUpdateSchema,
)
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._store import (
    Store,
    StoreDraft,
    StorePagedQueryResponse,
    StoreUpdate,
    StoreUpdateAction,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _StoreQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
):
    pass


class _StoreUpdateSchema(traits.ExpandableSchema, traits.VersionedSchema):
    pass


class _StoreDeleteSchema(traits.VersionedSchema, traits.ExpandableSchema):
    pass


class StoreService(abstract.AbstractService):
    """Stores let you model the context your customers shop in."""

    def get_by_id(self, id: str, *, expand: OptionalListStr = None) -> Store:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"stores/{id}", params=params, schema_cls=StoreSchema
        )

    def get_by_key(self, key: str, *, expand: OptionalListStr = None) -> Store:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"stores/key={key}", params=params, schema_cls=StoreSchema
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> StorePagedQueryResponse:
        """Stores let you model the context your customers shop in.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _StoreQuerySchema,
        )
        return self._client._get(
            endpoint="stores", params=params, schema_cls=StorePagedQueryResponseSchema
        )

    def create(self, draft: StoreDraft, *, expand: OptionalListStr = None) -> Store:
        """Stores let you model the context your customers shop in.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="stores",
            params=params,
            data_object=draft,
            request_schema_cls=StoreDraftSchema,
            response_schema_cls=StoreSchema,
        )

    def update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[StoreUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> Store:
        params = self._serialize_params({"expand": expand}, _StoreUpdateSchema)
        update_action = StoreUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"stores/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=StoreUpdateSchema,
            response_schema_cls=StoreSchema,
            force_update=force_update,
        )

    def update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[StoreUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> Store:
        params = self._serialize_params({"expand": expand}, _StoreUpdateSchema)
        update_action = StoreUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"stores/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=StoreUpdateSchema,
            response_schema_cls=StoreSchema,
            force_update=force_update,
        )

    def delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> Store:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _StoreDeleteSchema
        )
        return self._client._delete(
            endpoint=f"stores/{id}",
            params=params,
            response_schema_cls=StoreSchema,
            force_delete=force_delete,
        )

    def delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> Store:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _StoreDeleteSchema
        )
        return self._client._delete(
            endpoint=f"stores/key={key}",
            params=params,
            response_schema_cls=StoreSchema,
            force_delete=force_delete,
        )
